
class Bus:

    def attach(self, handler):
        pass

    def detach(self, handler):
        pass
    
    def publish(self, publishable):
        pass


class Commutator:

    def __init__(self):
        self.buses = []

    def attach(self, bus):
        self.buses.append(bus)

    def publish(self, publishable):
        for bus in self.buses:
            bus.publish(publishable)




from .mappers.commandmapper import CommandMapper, CommandRecord
from .mappers.eventmapper import EventMapper, EventRecord

from .bus import Bus, Commutator
from .command import ApplicationCommand
from .handler import handler
from .projection import Projection
from .projector import projector
from .query import ApplicationQuery
from .registry import Registry
from .service import ApplicationService




import pytest

from .example import (
    ExampleAggregate,
    ExampleEvent,
    ExampleValueObject,
    ExampleIdentity
)

@pytest.fixture
def value_object():
    return ExampleValueObject('a')
    
@pytest.fixture
def event(value_object):
    return ExampleEvent(value_object)


def test_aggregate_apply(event):
    a = ExampleAggregate(ExampleIdentity(id="A"))
    a.__apply__(event)
    assert a.mutated

def test_aggregate_idempotent(event):
    a = ExampleAggregate(ExampleIdentity(id="A"))
    a.__route__(event)
    a.__route__(event)

    assert a.counter == 1




from domainpy.application.bus import Bus


class MemoryBus(Bus):

    def __init__(self):
        self._handlers = []

    def attach(self, handler):
        self._handlers.append(handler)

    def detach(self, handler):
        self._handlers.remove(handler)

    def publish(self, publishable):
        for handler in self._handlers:
            handler.__handle__(publishable)
    


import pytest 

from .example import (
    ExampleValueObject,
    ExampleEvent
)

@pytest.fixture
def value_object():
    return ExampleValueObject('a')
    
@pytest.fixture
def event(value_object):
    return ExampleEvent(something=value_object)
    
def test_event_property_acccess(event, value_object):
    assert event.something.__class__ is ExampleValueObject
    assert event.something == value_object




from domainpy.domain.model.aggregate import AggregateRoot
from domainpy.domain.model.entity import DomainEntity
from domainpy.domain.model.event import DomainEvent
from domainpy.domain.model.value_object import ValueObject, Identity
from domainpy.domain.model.specification import Specification
from domainpy.domain.model.mutator import mutator

from domainpy.domain.repository import Repository

from domainpy.domain.service import DomainService



import inspect

class Dictable:
    
    @classmethod
    def __from_dict__(cls, dictionary):
        if hasattr(cls, '__annotations__'):
            attrs = cls.__dict__['__annotations__']
            
            kwargs0 = {}
            for k in attrs:
                expected_type = attrs[k]
                
                value = dictionary[k]
                
                if hasattr(expected_type, '__origin__') and expected_type.__origin__ in (tuple, list):
                    (expected_type0,) = expected_type.__args__
                    kwargs0[k] = expected_type.__origin__(
                        expected_type0.__from_dict__(v) 
                        for v in value
                    )
                elif expected_type in (str, int, float, bool):
                    if isinstance(value, expected_type):
                        kwargs0[k] = expected_type(value)
                    else:
                        raise TypeError(f'{k} should be type of {expected_type}, found {value.__class__.__name__}')
                elif isinstance(value, dict):
                    if Dictable in expected_type.mro():
                        kwargs0[k] = expected_type.__from_dict__(value)
                    else:
                        raise TypeError(f'{k} should be Dictable, found {expected_type}')
                else:
                    raise TypeError(f'{k} should be type of dict, found {expected_type}')
                
            return cls(**kwargs0)
        else:
            raise NotImplementedError(
                f'{cls.__name__} should have annotations'
            )
    
    def __to_dict__(self):
        if hasattr(self.__class__, '__annotations__'):
            attrs = self.__class__.__dict__['__annotations__']
            
            dictionary = {}
            for k in attrs:
                
                expected_type = attrs[k]
                
                value = self.__dict__[k]
                
                if hasattr(expected_type, '__origin__') and expected_type.__origin__ in (list, tuple):
                    (expected_type0,) = expected_type.__args__
                    if not all(isinstance(v, expected_type0) for v in value):
                        raise TypeError(f'{k} should be type {expected_type.__origin__.__class__.__name__}[{expected_type0}]')
                elif not isinstance(value, expected_type):
                    raise TypeError(f'{k} should be type {expected_type} by declaration, found {value.__class__.__name__}')
                
                if hasattr(expected_type, '__origin__') and expected_type.__origin__ in (list, tuple):
                    (expected_type0,) = expected_type.__args__

                    if issubclass(expected_type0, Dictable):
                        dictionary[k] = expected_type.__origin__(v.__to_dict__() for v in value)
                        
                elif isinstance(value, (str, int, float, bool)):
                    dictionary[k] = value
                elif isinstance(value, tuple):
                    dictionary[k] = tuple(v.__to_dict__() for v in value)
                elif isinstance(value, list):
                    dictionary[k] = list(v.__to_dict__() for v in value)
                elif isinstance(value, Dictable):
                    dictionary[k] = value.__to_dict__()
                else:
                    raise TypeError(f'{k} in {self.__class__.__name__} should be dictable or list/tuple (is {expected_type})')

            return dictionary
        else:
            raise KeyError(
                f'{self.__class__.__name__} should have annotations'
            )




class Specification:
    
    def is_satisfied_by(self, candidate):
        pass

    def and_(self, other):
        return ConjunctionSpecification(
            self,
            other
        )

    def or_(self, other):
        return DisjunctionSpecification(
            self,
            other
        )

    def not_(self):
        return NegationSpecification(self)

    def is_special_case_of(self, other):
        raise NotImplementedError(f'{self.__class__} should override is_special_case_of')

    def is_generalization_of(self, other):
        raise NotImplementedError(f'{self.__class__} should override is_generalization_of')

    def remainder_unsatisfied_by(self, candidate):
        if not self.is_satisfied_by(candidate):
            return self.__class__
        else:
            return None


class CompositeSpecification(Specification):
    
    def __init__(self, a: Specification, b: Specification):
        self.a = a
        self.b = b


class ConjunctionSpecification(CompositeSpecification):
    
    def is_satisfied_by(self, candidate):
        return (
            self.a.is_satisfied_by(candidate) 
            and self.b.is_satisfied_by(candidate)
        )

    def is_special_case_of(self, other: Specification):
        return (
            self.a.is_special_case_of(other)
            or self.b.is_special_case_of(other)
        )


class DisjunctionSpecification(CompositeSpecification):
    
    def is_satisfied_by(self, candidate):
        return (
            self.a.is_satisfied_by(candidate) 
            or self.b.is_satisfied_by(candidate)
        )

    def is_generalization_of(self, other: Specification):
        return (
            self.a.is_generalization_of(other)
            or self.b.is_generalization_of(other)
        )


class NegationSpecification(Specification):

    def __init__(self, spec: Specification):
        self.spec = spec
    
    def is_satisfied_by(self, candidate):
        return not self.spec.is_satisfied_by(candidate)



import boto3
import json

from domainpy.application.bus import Bus


class EventBridgeBus(Bus):

    def __init__(self, source):
        self.source = source

        self.cloudwatch_events = boto3.client('events')
        self.names = []

    def attach(self, handler):
        self.names.append(handler)

    def detach(self, handler):
        self.names.remove(handler)

    def publish(self, publishable):
        for name in self.names:
            self.cloudwatch_events.put_events(
                Entries=[
                    {
                        'Source': self.source,
                        'Detail': json.dumps(publishable.__to_dict__()),
                        'DetailType': publishable.__class__.__name__,
                        'EventBusName': name
                    }
                ]
            )
    



from domainpy.infrastructure.eventsourced.eventstore import EventStore, EventStream
from domainpy.infrastructure.eventsourced.managers.memory import MemoryEventRecordManager
from domainpy.infrastructure.eventsourced.managers.dynamodb import DynamoEventRecordManager
from domainpy.infrastructure.projectionsourced.managers.memory import MemoryProjectionRecordManager
from domainpy.infrastructure.buses.memory import MemoryBus
from domainpy.infrastructure.buses.eventbridge import EventBridgeBus




from domainpy.domain.model.aggregate import AggregateRoot
from domainpy.domain.model.event import DomainEvent
from domainpy.domain.model.value_object import ValueObject, Identity


class ExampleValueObject(ValueObject):
    something: str
    
    
class ExampleIdentity(Identity):
    pass

class ExampleEvent(DomainEvent):
    something: ExampleValueObject


class ExampleAggregate(AggregateRoot):
    
    def __init__(self, *args, **kwargs):
        super(ExampleAggregate, self).__init__(*args, **kwargs)

        self.counter = 0
    
    def mutate(self, event):
        if(event.__class__ is ExampleEvent):
            self.mutated = True
            self.counter = self.counter + 1
    



class EventParameterIsNotValueObjectError(Exception):
    pass


class MutatorNotFoundError(Exception):
    pass


class SingleMutatorBrokenError(Exception):
    pass


class ValueObjectIsNotSerializable(Exception):
    pass




from functools import update_wrapper, partial

from domainpy.application.command import ApplicationCommand
from domainpy.application.service import ApplicationService
from domainpy.application.exceptions import (
    HandlerNotFoundError,
    MessageSingleHandlerBroken
)


class handler:
    
    def __init__(self, func):
        update_wrapper(self, func)
        
        self.func = func
        
        self._messages = dict()
        
    def __get__(self, obj, objtype):
        """Support instance methods."""
        return partial(self.__call__, obj)
        
    def __call__(self, service, message):
        if(message.__class__ not in self._messages):
            #raise HandlerNotFoundError((message.__class__.__name__ + " in " + service.__class__.__name__))
            return
        
        results = []
        
        handlers = self._messages.get(message.__class__, [])
        for h in handlers:
            results.append(h(service, message))
            
        return results
            
    def command(self, command_type: type):
        def inner_function(func):
            
            if command_type in self._messages:
                raise MessageSingleHandlerBroken(f'handler already defined for {command_type}')
                
            self._messages.setdefault(command_type, set()).add(func)
            
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)
            return wrapper
        return inner_function
        
    def query(self, query_type: type):
        def inner_function(func):
            
            if query_type in self._messages:
                raise MessageSingleHandlerBroken(f'handler already defined for {query_type}')
            
            self._messages.setdefault(query_type, set()).add(func)
            
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)
            return wrapper
        return inner_function
        
    def event(self, event_type: type):
        def inner_function(func):
            
            self._messages.setdefault(event_type, set()).add(func)
            
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)
            return wrapper
        return inner_function



import decimal
import boto3
from boto3.dynamodb.conditions import Key, Attr

from domainpy.application.mappers.eventmapper import EventRecord
from domainpy.infrastructure.eventsourced.recordmanager import (
    EventRecordManager,
    Session
)

class DynamoEventRecordManager(EventRecordManager):

    def __init__(self, table_name):
        dynamodb = boto3.resource('dynamodb')

        self.table = dynamodb.Table(table_name)

    def session(self):
        return DynamoSession(self)

    def find(self, stream_id: str):
        query = self.table.query(
            KeyConditionExpression=Key('stream_id').eq(stream_id)
        )
        return tuple([
            EventRecord(
                stream_id=deserialize(i['stream_id']),
                number=deserialize(i['number']),
                topic=deserialize(i['topic']),
                version=deserialize(i['version']),
                timestamp=deserialize(i['timestamp']),
                message=deserialize(i['message']),
                payload=deserialize(i['payload'])
            )
            for i in query['Items']
        ])


class DynamoSession(Session):

    def __init__(self, record_manager):
        self.writer = record_manager.table

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        #self.writer.__exit__(*args, **kwargs)
        pass

    def append(self, event_record: EventRecord):
        if event_record is None:
            raise TypeError('event_record cannot be None')
        
        print('WRITING', event_record)
        self.writer.put_item(
            Item={
                'stream_id': event_record.stream_id,
                'number': event_record.number,
                'topic': event_record.topic,
                'version': event_record.version,
                'timestamp': event_record.timestamp,
                'message': event_record.message,
                'payload': event_record.payload
            },
            ConditionExpression=(
                Attr('stream_id').not_exists()
                & Attr('number').not_exists()
            )
        )
        print('WRITTEN')

    def commit(self):
        pass

    def rollback(self):
        pass
    

def deserialize(obj):
    if isinstance(obj, list):
        for i in xrange(len(obj)):
            obj[i] = deserialize(obj[i])
        return obj
    elif isinstance(obj, dict):
        for k in obj.keys():
            obj[k] = deserialize(obj[k])
        return obj
    elif isinstance(obj, decimal.Decimal):
        if obj % 1 == 0:
            return int(obj)
        else:
            return float(obj)
    else:
        return obj



from domainpy.application import (
    Bus,
    Commutator,
    ApplicationService,
    ApplicationCommand,
    ApplicationQuery,
    Projection,
    CommandMapper,
    EventMapper,
    handler,
    projector,
)

from domainpy.domain import (
    AggregateRoot,
    DomainEntity,
    DomainEvent,
    ValueObject,
    Identity,
    Repository,
    DomainService,
    mutator
)

from domainpy.infrastructure import (
    EventStore,
    EventStream,
    MemoryEventRecordManager,
    DynamoEventRecordManager,
    MemoryProjectionRecordManager,
    MemoryBus,
    EventBridgeBus
)




class ProjectionRecordManager:
    
    def session(self):
        raise NotImplementedError(f'{self.__class__.__name__} must override session method')
        
    def get_item(self, attributes):
        raise NotImplementedError(f'{self.__class__.__name__} must override get_item method')


class Session:

    def put_item(self, item):
        raise NotImplementedError(f'{self.__class__.__name__} must override put_item method')

    def update_item(self, key, item):
        raise NotImplementedError(f'{self.__class__.__name__} must override update_item method')

    def delete_item(self, key):
        raise NotImplementedError(f'{self.__class__.__name__} must override delete_item method')



from domainpy.utils.constructable import Constructable
from domainpy.utils.immutable import Immutable
from domainpy.utils.dictable import Dictable

class ApplicationCommand(Constructable, Immutable, Dictable):

    def __init__(self, *args, **kwargs):
        self.__dict__.update({
            '__message__': 'command'
        })
        
        super(ApplicationCommand, self).__init__(*args, **kwargs)




from datetime import datetime

from domainpy.domain.model.value_object import Identity
from domainpy.domain.model.event import DomainEvent


class AggregateRoot:
    
    def __init__(self, id: Identity):
        if not isinstance(id, Identity):
            raise TypeError('id should be type of Identity')
        
        self.__id__ = id
        
        self.__version__ = 0
        self.__changes__ = [] # New events
        self.__seen__ = [] # Idempotent
    
    def __apply__(self, event: DomainEvent):
        self.__stamp__(event)
        self.__route__(event)
        
        self.__changes__.append(event)
    
    def __stamp__(self, event: DomainEvent):
        event.__dict__.update({
            '__aggregate_id__': f'{self.__id__.id}:{self.__class__.__name__}',
            '__number__': self.__version__ + 1,
            '__version__': 1,
            '__timestamp__': str(datetime.now())
        })
        
    def __route__(self, event: DomainEvent):
        if event not in self.__seen__:
            self.__version__ = event.__number__
            self.__seen__.append(event)

            self.mutate(event)
    
    def mutate(self, event: DomainEvent):
        pass
